/*
 *    Database.java
 *    Copyright (C) 1999 Pierre Geurts
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import java.util.*;
import java.io.*;
import java.net.*;

public class Database{
  /* the name of the database */
  String name;
  /* the filename from which the database has been loaded */
  String filename;
  /* the set of objects */
  ObjectSet objects=new ObjectSet();
  /* the vector of atttributes */
  Vector attributes=new Vector();

/* method to add an object */
  public int addObjToDb(Obj o) {
    objects.addElement(o);
    return objects.size-1;
  }

/* method to add an attribute */
  public int addAttToDb(Attribute att){
    attributes.addElement(att);
    return attributes.size()-1;
  }
  
/* a short description of the database */
  public String toString(){
    StringBuffer s=new StringBuffer();
    s.append(name);
    s.append(" o(");
    s.append(objects.size);
    s.append(") a(");
    s.append(attributes.size());
    s.append(")");
    return s.toString();
  }
  
/* to "pretty" print the content of the database */
  public void printDb() {
    System.out.println(toString());
    System.out.print("Object  ");
    for (int i=0; i<attributes.size(); i++) {
      System.out.print(((Attribute)attributes.elementAt(i)).toString() + "  ");
    }
    System.out.println("");
    for (int i=0; i<objects.size; i++) {
      Obj o=(Obj)objects.elementAt(i);
      System.out.print(o.toString() + "  ");
      for (int j=0; j<attributes.size(); j++) {
	System.out.print(o.attributesvalues.elementAt(j).toString());
	System.out.print("  ");
      }
      System.out.println("");
    }  
  }

/* utilities */
  private static void skipBlankLines(StreamTokenizer stt) throws IOException {
    while (stt.ttype==stt.TT_EOL)
	stt.nextToken();
  }

/* loading of a database from a file */
  /* return 1 if there is an error, 0 otherwise */
  public int loadDb (URL u) {
    boolean objectname=false;
    int objectnameindex=0;
    filename=u.toString();
    try {
      //      URL u=new URL(fn);
      URLConnection urlConnection=u.openConnection();
      urlConnection.connect();
      
      InputStream in=urlConnection.getInputStream();
      BufferedReader r=new BufferedReader(new InputStreamReader(in));
      StreamTokenizer stt=new StreamTokenizer(r);
      /* some streamtokenizer inititialisations*/
      stt.lowerCaseMode(false);
      stt.parseNumbers();
      stt.eolIsSignificant(true);
      stt.commentChar(';');
      stt.wordChars('$','$');
      stt.wordChars('_','_');
      /* file reading */
      stt.nextToken();
      skipBlankLines(stt);
      
      /**************************
       * read the database name *
       **************************/
      if (stt.ttype==stt.TT_WORD)
	name=stt.sval;
      else {
	System.out.println("the database name must be a symbol");
	return 1;
      }
      stt.nextToken();
      skipBlankLines(stt);

      /*******************************
       * read attributes declaration *
       *******************************/
      int nreadvalue=0;
      int natt=0;
      while (stt.ttype!=stt.TT_EOL) {
	String attname;
	String atttype;
	if (stt.ttype!=stt.TT_WORD) {
	  System.out.println("error in attribute declaration");
	  return 1;
	}
	attname=stt.sval;
	stt.nextToken();
	if (stt.ttype!=stt.TT_WORD) {
	  System.out.println("error in attribute declaration");
	  return 1;
	}
	atttype=stt.sval;
	if (atttype.equals("name") || atttype.equals("NAME")) {
	  if (!objectname) {
	    objectname=true;
	    objectnameindex=nreadvalue;
	  } else {
	    System.out.println("error in attribute declaration");
	    return 1;
	  }
	} else {
	  if (atttype.equals("symbolic") || atttype.equals("SYMBOLIC"))
	    new SymbAtt(attname,this);
	  else if (atttype.equals("numerical") || atttype.equals("NUMERICAL"))
	    new NumAtt(attname,this);
	  else {
	    System.out.println("not recognised attribute type");
	    return 1;
	  }
	  natt++;
	}
	nreadvalue++;
	stt.nextToken();
      }
      System.out.println("" + natt + " attributes have been read");
      System.out.println(attributes.toString());
      
      /**************************
       * read objects definitions *
       **************************/
      
      int exactnatt=natt;
      int nobject=1;
      while(stt.ttype!=stt.TT_EOF) {
	skipBlankLines(stt);
	/* number of readed values*/
	nreadvalue=0;
	/* number of readed attributes (as nreadvalue but 
	   without the object name) */
	natt=0;
	Obj o;
	if (objectname)
	  o=new Obj("",this);
	else
	  o=new Obj("O" + nobject,this);
	nobject++;
	
	/* read attributes values */
	while(stt.ttype!=stt.TT_EOL && stt.ttype!=stt.TT_EOF) {
	  if (objectname==true && 
	      objectnameindex==nreadvalue && 
	      nreadvalue==objectnameindex) {
	    if (stt.ttype==stt.TT_WORD) {
	      o.name=stt.sval;
	      nreadvalue++;
	    } else {
	      System.out.println("Object name must be a string");
	      return 1;
	    }
	  } else {
	    if (((Attribute)attributes.elementAt(natt)).ReadAttValue(o,stt)
		==1) {
	      System.out.println("objet " + nobject + " value " + nreadvalue);
	      return 1;
	    }
	    natt++;
	    nreadvalue++;
	  }
	  stt.nextToken();
	}
	if (natt<exactnatt) {
	  System.out.println("Attribute values are missing");
	  return 1;
	}
	skipBlankLines(stt);
      }
      System.out.println("" + (nobject-1) + " objects have been read");
      // System.out.println(objects.toString());
      in.close();
    }
    catch (Exception e){
      System.err.println(e);
    }
    return 0;
  }

/* a method to find an attribute by its name */
  public Attribute findAttribute(String name) {
    int as=attributes.size();
    for (int i=0; i<as; i++) {
      Attribute a=(Attribute)attributes.elementAt(i);
      if (name.equals(a.name))
	return a;
    }
    return null;
  }

/* a method to find all attributes except one   */
  public Vector getCandidateAttributes(String goal) {
    int as=attributes.size();
    Attribute a;
    Vector result=new Vector();
    for (int i=0; i<as; i++) {
      a=(Attribute)attributes.elementAt(i);
      if (!goal.equals(a.name)) {
	result.addElement(a);
      }
    }
    return result;
  }

/* a method to find a object by its name */
  public Obj findObject(String name) {
    int os=objects.size;
    for (int i=0; i<os; i++) {
      Obj o=(Obj)objects.elementAt(i);
      if (name.equals(o.name))
	return o;
    }
    return null;
  }
}
